% Causal bootstrapping for front-door deconfounding with discrete, observed
% mediator Z and discrete cause variable Y, with arbitrary multivariate
% effect X. Deconfounds the given data in variables X,Y,Z under the
% mediated, unobserved confounding causal structural model where
% U->Y->Z->X<-U (U represents any unobserved confounding path).
% Inputs: X,Y,Z - joint samples for X (NxD matrix where N is the
%                 number of samples), Y (Nx1 vector) and Z (Nx1 vector)
%         yr,zr - column vectors of values that Y and U can take
%             M - number of bootstrap samples to produce
% Outputs:
%      Xw,Yw,Zw - joint deconfounded samples for X,Y,Z
%
% (cc) Attribution-ShareAlike 3.0, Max Little, 2020. If you use this code,
% please cite M.A. Little, R. Badawy, "Causal Bootstrapping", 2019,
% arXiv:1910.09648

function [Xw,Yw,Zw] = cb_frontdoor(X,Y,Z,yr,zr,M)

% Step 1. Estimate f(z,y), f(y) and f(z|y)
N = size(X,1);
Nzy = histcnd([Z Y],{zr,yr});
pzy = Nzy/N;
py = sum(pzy,1)';
pz_y = pzy./repmat(py',2,1);

Ns = round(py*M);
i = NaN(M,1);
Yw = NaN(M,1);

% Step 2. For each y in range of values of Y variable
k = 1;
for y = 1:length(yr)

    % Step 3. Resample indices
    idx = sub2ind(size(pz_y),Z,Y);
    v = pz_y(Z,y)./pz_y(idx)/N;
    l = k+Ns(y)-1;
    i(k:l) = rand_cat_fast(v,Ns(y));
    
    % Set resampled intervention data Y=y
    Yw(k:l) = yr(y);
    
    k = k + Ns(y);
end

% 4. Bootstrapped Z,X data
Zw = Z(i);
Xw = X(i,:);
