clear all;
close all;

% Simulate environments:
% 1 - Confounded training data
% 2 - Confounded test data
% 3 - Non-confounded test data

Ne = 3;

% Structural causal model parameters
% Feature data X, prediction target class Y, confounder U, mediator Z

% (b) Background brightness confounded MNIST data

fprintf('-- Background-MNIST continuous unmeasured confounding\n');

% For reproducibility
randn('state',102);
rand('state',1);

clear X Y Z U;

datadir = '.';

% Load image data as Nw*Nh by Nimg matrix
Nw = 28;
Nh = 28;
data = [];
digits = [2 6];
Ndigits = length(digits);
labels = [];
for i = 1:length(digits)
    fid = fopen(sprintf('%s/data%d',datadir,digits(i)),'rb');
    digitdata = fread(fid,[Nw*Nh inf],'uchar');
    Nchars = size(digitdata,2);
    data = [data digitdata];
    labels = [labels ones(1,Nchars)*digits(i)];
    fclose(fid);
end

data = data';
labels = labels';
N = size(data,1);

% U->Y strength (confounding)
quy = [0.80 0.80 0.50];

% Confounding background brightness adjustment
bright = 100;

% Y->Z strength (primary effect)
qyz = [0.95 0.95 0.95];

% Repetitions to get accuracy variance
R = 10;
E = NaN(R,4);
for r = 1:R

% Randomly permute image data
i = randperm(N);
data = data(i,:);
labels = labels(i);

% Select input image data as the basis for each environment
Nie = round(N/Ne);
ie = cell(Ne,1);
e = 1;
i = 1;
while (i <= N)
    il = i:(i+Nie-1);
    il(il > N) = [];
    ie{e} = il';
    i = i + Nie;
    e = e + 1;
end

for e = 1:Ne
    
    pz_y = [qyz(e) 1-qyz(e); 1-qyz(e) qyz(e)];  % P(Z|Y)

    % Draw (continuous) confounder variable U
    M = round(size(ie{e},1)*0.90);
    usig = 5;
    umu = 0;
    U{e} = usig*randn(M,1)+umu;

    % Draw (binary) target classification variable Y
    q = log((1-quy(e))/quy(e));
    pyy = 1./(1+exp(q*U{e}));
    u = rand(M,1);
    Y{e} = double(u <= pyy)+1;

    % Draw (binary) mediator variable Z
    u = rand(M,1);
    p = pz_y(2,Y{e})';
    Z{e} = double(u<=p)+1;
    
    % "Draw" image data variable X selected by Z, modified by U
    clear idx;
    for k = 1:Ndigits
        idx{k} = intersect(find(labels == digits(k)),ie{e});
    end
    X{e} = NaN(M,Nw*Nh);
    j = ones(Ndigits,1);
    for n = 1:M
        ilist = idx{Z{e}(n)};
        i = ilist(j(Z{e}(n)));
        j(Z{e}(n)) = j(Z{e}(n)) + 1;
        x = data(i,:) + repmat((0.5*tanh(0.2*U{e}(n))+0.5)*bright,1,Nw*Nh);
        x(x > 255) = 255;
        X{e}(n,:) = x;
    end

end

% Apply front-door deconfounding
[Xw,Yw,Zw] = cb_frontdoor(X{1},Y{1},Z{1},1:2,1:2,N);

% KNN predictions
K = 5;

Yp = knn(X{1},X{2},Y{1},K);
E_conf_conf = mean(Yp == Y{2});

Yp = knn(X{1},X{3},Y{1},K);
E_conf_nc = mean(Yp == Y{3});

Yp = knn(Xw,X{2},Yw,K);
E_fd_conf = mean(Yp == Y{2});

Yp = knn(Xw,X{3},Yw,K);
E_fd_nc = mean(Yp == Y{3});

E(r,:) = [E_conf_conf E_conf_nc E_fd_conf E_fd_nc];
fprintf('Repetition: %d\n',r);

end

muE = mean(E);
stdE = std(E);

fprintf('MNIST KNN (K=%d) prediction accuracies:\n', K);
fprintf('Confound train data -> confound test:       %0.2f(%0.2f)\n', muE(1), stdE(1));
fprintf('Confound train data -> non-confound test:   %0.2f(%0.2f)\n', muE(2), stdE(2));
fprintf('Deconfound train data -> confound test:     %0.2f(%0.2f)\n', muE(3), stdE(3));
fprintf('Deconfound train data -> non-confound test: %0.2f(%0.2f)\n', muE(4), stdE(4));

