clear all;
close all;

% Simulate environments:
% 1 - Confounded training data
% 2 - Confounded test data
% 3 - Non-confounded test data

Ne = 3;

% Structural causal model parameters
% Feature data X, prediction target class Y, confounder U, mediator Z

fprintf('-- Gaussian mixture discrete unmeasured confounding\n');

% For reproducibility
randn('state',92);
rand('state',0);

clear X Y Z U;

N = 1e3;

% U->Y strength (confounding)
quy = [0.98 0.98 0.50];

% Y->Z strength (primary effect)
qyz = [0.90 0.90 0.90];

% Z->X strength (mediator effect)
% qzx = [1.0 1.0 1.0];   % Strong separation
% qzx = [0.8 0.8 0.8];   % Large separation
qzx = [0.5 0.5 0.5];   % Medium separation

% U->X strength (modifier effect)
qux = [1.0 1.0 1.0];   % Strong separation
% qux = [0.8 0.8 0.8];   % Strong separation
% qux = [0.2 0.2 0.2];   % Strong separation

% pu = [0.85 0.15];                % Confounder marginal P(U)
pu = [0.5 0.5];                % Confounder marginal P(U)

% Repetitions to get accuracy variance
R = 10;
E = NaN(R,4);
for r = 1:R

% Run through each environment
for e = 1:Ne
    
    py_u = [quy(e) 1-quy(e); 1-quy(e) quy(e)];  % P(Y|U)
    pyu = py_u.*repmat(pu,2,1);   % P(Y,U)
    py = sum(pyu,2);               % P(Y)

    pz_y = [qyz(e) 1-qyz(e); 1-qyz(e) qyz(e)];  % P(Z|Y)
    pzy = pz_y.*repmat(py',2,1);    % P(Z,Y)
    pz = sum(pzy,2);                % P(Z)

    % Draw (binary) confounder variable U
    u = rand(N,1);
    U{e} = double(u<=pu(2))+1;

    % Draw (binary) target variable Y
    u = rand(N,1);
    p = py_u(2,U{e})';
    Y{e} = double(u<=p)+1;

    % Draw (binary) mediator variable Z
    u = rand(N,1);
    p = pz_y(2,Y{e})';
    Z{e} = double(u<=p)+1;

    % Setup means of Gaussian feature variables P(X|Z,U)
    mux = zeros(N,2);
    i = (Z{e}==1);
    mux(i,1) = mux(i,1) + 3*qzx(e);
    mux(~i,1) = mux(~i,1) - 3*qzx(e);
    i = (U{e}==1);
    mux(i,2) = mux(i,2) + 3*qux(e);
    mux(~i,2) = mux(~i,2) - 3*qux(e);

    % Draw Gaussian feature variables
    X{e} = mux + randn(N,2);

end

% LDA predictions
[Xw,Yw,Zw] = cb_frontdoor(X{1},Y{1},Z{1},1:2,1:2,N);
lda_conf = ldatrain(X{1},Y{1});
lda_fd = ldatrain(Xw,Yw);
E_conf_conf = ldapredict(X{2},Y{2},lda_conf);
E_conf_nc = ldapredict(X{3},Y{3},lda_conf);
E_fd_conf = ldapredict(X{2},Y{2},lda_fd);
E_fd_nc = ldapredict(X{3},Y{3},lda_fd);
E(r,:) = [E_conf_conf E_conf_nc E_fd_conf E_fd_nc];
fprintf('Repetition: %d\n',r);

end

muE = mean(E);
stdE = std(E);

fprintf('Gaussian mixture LDA prediction accuracies:\n');
fprintf('Confound train data -> confound test:       %0.2f(%0.2f)\n', muE(1), stdE(1));
fprintf('Confound train data -> non-confound test:   %0.2f(%0.2f)\n', muE(2), stdE(2));
fprintf('Deconfound train data -> confound test:     %0.2f(%0.2f)\n', muE(3), stdE(3));
fprintf('Deconfound train data -> non-confound test: %0.2f(%0.2f)\n', muE(4), stdE(4));


