% Causal bootstrapping for back-door deconfounding with discrete, observed
% confounder U and discrete cause variable Y, with arbitrary multivariate
% effect X. Deconfounds the given data in variables X,Y,U under the
% simple confounding causal structural model where U->Y->X<-U.
% Inputs: X,Y,U - joint samples for X (NxD matrix where N is the
%                 number of samples), Y (Nx1 vector) and U (Nx1 vector)
%         yr,ur - column vectors of values that Y and U can take
%             M - number of bootstrap samples to produce
% Outputs:
%      Xw,Yw,Uw - joint deconfounded samples for X,Y,U
%
% (cc) Attribution-ShareAlike 3.0, Max Little, 2020. If you use this code,
% please cite M.A. Little, R. Badawy, "Causal Bootstrapping", 2019,
% arXiv:1910.09648

function [Xw,Yw,Uw] = cb_backdoor(X,Y,U,yr,ur,M)

% Step 1. Estimate f(u,y), f(y) and f(u|y)
N = size(X,1);
K = size(yr,1);
Nyu = histcnd([Y U],{yr,ur});
pyu = Nyu/N;
pu = sum(pyu,1)';
py = sum(pyu,2);
py_u = pyu./repmat(pu',2,1);

Ns = round(py(1:K-1)*M);
Ns(K) = M-sum(Ns(1:K-1));

i = NaN(M,1);
Yw = NaN(M,1);

% Step 2. For each y in range of values of Y variable
k = 1;
for y = 1:length(yr)

    % Step 3. Resample indices
    v = (Y==yr(y))./py_u(y,U)'/N;
    l = k+Ns(y)-1;
    i(k:l) = rand_cat_fast(v,Ns(y));
    
    % Set resampled intervention data Y=y
    Yw(k:l) = yr(y);

    k = k + Ns(y);
end

% 4. Bootstrapped U,X data
Uw = U(i);
Xw = X(i,:);
