clear all;
close all;

% Simulate environments:
% 1 - Confounded training data
% 2 - Confounded test data
% 3 - Non-confounded test data

Ne = 3;

% Structural causal model parameters: U->Y->X<-U
% Feature data X, prediction target class Y, discrete confounder U

% For reproducibility
randn('state',92);
rand('state',0);    

fprintf('-- Gaussian mixture discrete measured confounding\n');

clear X Y U;

N = 1e3;

% U->Y strength (confounding)
qyu = [0.95 0.95 0.5];

% Y->X strength (main effect)
qyx = [0.5 0.5 0.5];   % Medium separation

% U->X strength (modifier effect)
qux = [0.8 0.8 0.8];   % Strong separation

pu = [0.85 0.15];                % Confounder marginal P(U)


% Repetitions to get accuracy variance
% R = 1;
R = 10;
E = NaN(R,4);
for r = 1:R

% Run through each environment
for e = 1:Ne
    
    py_u = [qyu(e) 1-qyu(e); 1-qyu(e) qyu(e)];   % P(Y|U) (correlated)
    % py_u = [1-qyu(e) qyu(e); qyu(e) 1-qyu(e)]   % P(Y|U) (anti-correlated)
    pyu = repmat(pu',1,2).*py_u;     % P(Y,U)
    py = sum(pyu,1)';                % P(Y)

    % Draw (binary) confounder variable U
    u = rand(N,1);
    U{e} = double(u<=pu(2))+1;

    % Draw (binary) target classification variable Y
    u = rand(N,1);
    p = py_u(U{e},2);
    Y{e} = double(u<=p)+1;

    % Setup means of Gaussian feature variables P(X|Y,U)
    mux = zeros(N,2);
    i = (Y{e}==1);
    mux(i,1) = mux(i,1) + 3*qyx(e);
    mux(~i,1) = mux(~i,1) - 3*qyx(e);
    i = (U{e}==1);
    mux(i,2) = mux(i,2) + 3*qux(e);
    mux(~i,2) = mux(~i,2) - 3*qux(e);

    % Draw Gaussian feature variables
    X{e} = mux + randn(N,2);

end

% LDA predictions
[Xw,Yw,Uw] = cb_backdoor(X{1},Y{1},U{1},unique(Y{1}),unique(U{1}),N);
lda_conf = ldatrain(X{1},Y{1});
lda_fd = ldatrain(Xw,Yw);
E_conf_conf = ldapredict(X{2},Y{2},lda_conf);
E_conf_nc = ldapredict(X{3},Y{3},lda_conf);
E_fd_conf = ldapredict(X{2},Y{2},lda_fd);
E_fd_nc = ldapredict(X{3},Y{3},lda_fd);
E(r,:) = [E_conf_conf E_conf_nc E_fd_conf E_fd_nc];
fprintf('Repetition: %d\n',r);

end

muE = mean(E);
stdE = std(E);

fprintf('Gaussian mixture LDA prediction accuracies:\n');
fprintf('Confound train data -> confound test:       %0.2f(%0.2f)\n', muE(1), stdE(1));
fprintf('Confound train data -> non-confound test:   %0.2f(%0.2f)\n', muE(2), stdE(2));
fprintf('Deconfound train data -> confound test:     %0.2f(%0.2f)\n', muE(3), stdE(3));
fprintf('Deconfound train data -> non-confound test: %0.2f(%0.2f)\n', muE(4), stdE(4));
